"""
implementation of the PWC-DC network for optical flow estimation by Sun et al., 2018
Jinwei Gu and Zhile Ren
"""

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import os
os.environ['PYTHON_EGG_CACHE'] = 'tmp/' # a writable directory 
import numpy as np
import math
import pdb
import time

from .submodule import pspnet_s, pspnet, conv
from .conv4d import sepConv4d, butterfly4D
from ..base_model.base_model import BaseModel


class VCNLoss(nn.Module):
    def __init__(self, args):
        super(VCNLoss, self).__init__()
        self.maxdisp = args.maxdisp
        self.fac = args.fac
        self.bs = args.train_batch_size

        self.warpx = WarpModule()

    def forward(self, preds, inputs):
        flowl0 = inputs['flows'][:, 0].clone()
        mask = (inputs['valids'][:, 0, 0] == 1) & (flowl0[:,0].abs() < self.maxdisp) & (flowl0[:,1].abs() < (self.maxdisp//self.fac))
        mask = mask.detach().clone()

        im = inputs['images']
        im = im.view(im.shape[0]*im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        # out-of-range loss
        im_warp, _ = self.warpx(im[self.bs:], flowl0[:,:2])
        occ_mask = (im_warp - im[:self.bs]).norm(dim=1)>0.3

        m = 64.0
        loss = 0
        ws = [0.25, 0.25, 0.25, 0.5, 1.0]
        for i in range(len(ws)):
            loss += ws[i]*torch.norm((preds['flow_preds'][i]*m-flowl0[:, :2]), 0, 1)[mask].mean()
            m /= 2

        im_size = flowl0.shape[2:4]

        m = 32.0
        up_flows = preds['up_flows_preds']
        for i in range(len(up_flows)):
            up_flows[i] = F.interpolate(up_flows[i], [im_size[0], im_size[1]], mode='bilinear', align_corners=False)*m
            m /= 2
        up_flows.insert(0, 0)

        oors = preds['oors_preds']
        for i in range(len(oors)):
            oors[i] = F.interpolate(oors[i][:,np.newaxis], [im_size[0], im_size[1]], mode='bilinear', align_corners=False)[:,0]

        m = 64
        for i in range(len(oors)):
            loss += self.get_oor_loss(flowl0[:,:2]-up_flows[i], oors[i], (m * preds['flow_reg_maxs_preds'][i]), occ_mask)
            m /= 2

        return loss

    def get_oor_loss(self, flowl0, oor3, maxdisp, occ_mask):
        """ 
        return out-of-range loss
        """
        oor3_gt = (flowl0.abs() > maxdisp).detach() #  (8*self.md[3])
        oor3_gt = (((oor3_gt.sum(1)>0) + occ_mask)>0).float()  # oor, or occluded
        weights = oor3_gt.sum().float()/(oor3_gt.shape[0]*oor3_gt.shape[1]*oor3_gt.shape[2])
        weights = oor3_gt * (1-weights) + (1-oor3_gt) * weights
        loss_oor3 = F.binary_cross_entropy_with_logits(oor3,oor3_gt,size_average=True, weight=weights)
        return loss_oor3


class flow_reg(nn.Module):
    """
    Soft winner-take-all that selects the most likely diplacement.
    Set ent=True to enable entropy output.
    Set maxdisp to adjust maximum allowed displacement towards one side.
        maxdisp=4 searches for a 9x9 region.
    Set fac to squeeze search window.
        maxdisp=4 and fac=2 gives search window of 9x5
    """
    def __init__(self, ent=False, maxdisp = int(4), fac=1):
        super(flow_reg, self).__init__()
        self.ent = ent
        self.md = maxdisp
        self.fac = fac
        self.truncated = True
        self.wsize = 3  # by default using truncation 7x7

        self.create_flow([1, 1, 1])

        self.pool3d = nn.MaxPool3d((self.wsize*2+1,self.wsize*2+1,1),stride=1,padding=(self.wsize,self.wsize,0))

    def create_flow(self, size):
        B,W,H = size
        flowrangey = range(-self.md,self.md+1)
        flowrangex = range(-int(self.md//self.fac),int(self.md//self.fac)+1)
        meshgrid = np.meshgrid(flowrangex,flowrangey)
        flowy = np.tile( np.reshape(meshgrid[0],[1,2*self.md+1,2*int(self.md//self.fac)+1,1,1]), (B,1,1,H,W) )
        flowx = np.tile( np.reshape(meshgrid[1],[1,2*self.md+1,2*int(self.md//self.fac)+1,1,1]), (B,1,1,H,W) )
        self.flowx = torch.Tensor(flowx)
        self.flowy = torch.Tensor(flowy)

    def forward(self, x):
        b,u,v,h,w = x.shape
        oldx = x

        if b != self.flowx.shape[0] or h != self.flowx.shape[3] or w != self.flowx.shape[4]:
            self.create_flow((b, w, h))

        if self.truncated:
            # truncated softmax
            x = x.view(b,u*v,h,w)

            idx = x.argmax(1)[:,np.newaxis]
            if x.is_cuda:
                mask = Variable(torch.cuda.HalfTensor(b,u*v,h,w)).fill_(0)
            else:
                mask = Variable(torch.FloatTensor(b,u*v,h,w)).fill_(0)
            mask.scatter_(1,idx,1)
            mask = mask.view(b,1,u,v,-1)
            mask = self.pool3d(mask)[:,0].view(b,u,v,h,w)

            ninf = x.clone().fill_(-np.inf).view(b,u,v,h,w)
            x = torch.where(mask.byte(),oldx,ninf)
        else:
            self.wsize = (np.sqrt(u*v)-1)/2

        b,u,v,h,w = x.shape
        x = F.softmax(x.view(b,-1,h,w),1).view(b,u,v,h,w)
        outx = torch.sum(torch.sum(x*self.flowx.to(x.device),1),1,keepdim=True)
        outy = torch.sum(torch.sum(x*self.flowy.to(x.device),1),1,keepdim=True)

        if self.ent:
            # local
            local_entropy = (-x*torch.clamp(x,1e-9,1-1e-9).log()).sum(1).sum(1)[:,np.newaxis]
            if self.wsize == 0:
                local_entropy[:] = 1.
            else:
                local_entropy /= np.log((self.wsize*2+1)**2)

            # global
            x = F.softmax(oldx.view(b,-1,h,w),1).view(b,u,v,h,w)
            global_entropy = (-x*torch.clamp(x,1e-9,1-1e-9).log()).sum(1).sum(1)[:,np.newaxis]
            global_entropy /= np.log(x.shape[1]*x.shape[2])
            return torch.cat([outx,outy],1),torch.cat([local_entropy, global_entropy],1)
        else:
            return torch.cat([outx,outy],1),None


class WarpModule(nn.Module):
    """
    taken from https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
    """
    def __init__(self):
        super(WarpModule, self).__init__()
        self.create_grid([1, 1, 1])

    def create_grid(self, size):
        B,W,H = size
        # mesh grid 
        xx = torch.arange(0, W).view(1,-1).repeat(H,1)
        yy = torch.arange(0, H).view(-1,1).repeat(1,W)
        xx = xx.view(1,1,H,W).repeat(B,1,1,1)
        yy = yy.view(1,1,H,W).repeat(B,1,1,1)
        
        # self.register_buffer('grid',torch.cat((xx,yy),1).float())
        self.grid = torch.cat((xx,yy),1).float()

    def forward(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        if B != self.grid.shape[0] or H != self.grid.shape[2] or W != self.grid.shape[3]:
            self.create_grid((B, W, H))

        vgrid = self.grid.to(flo.device) + flo

        # scale grid to [-1,1] 
        vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:]/max(W-1,1)-1.0
        vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:]/max(H-1,1)-1.0

        vgrid = vgrid.permute(0,2,3,1)        
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = ((vgrid[:,:,:,0].abs()<1) * (vgrid[:,:,:,1].abs()<1)) >0
        return output*mask.unsqueeze(1).float(), mask


class VCNSmall(BaseModel):
    """
    VCN.
    md defines maximum displacement for each level, following a coarse-to-fine-warping scheme
    fac defines squeeze parameter for the coarsest level
    """

    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn_small-chairs-2a5f72ba.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn_small-things-e5a44ada.ckpt'
    }

    def __init__(self, args, md=[4,4,4,4,4]):
        super(VCNSmall, self).__init__(
            args=args,
            loss_fn=VCNLoss(args),
            output_stride=64)
        self.md = md
        self.fac = self.args.fac
        use_entropy = True
        withbn = True
        
        ## pspnet
        self.pspnet = pspnet_s(is_proj=False)

        ## Volumetric-UNet
        fdima1 = 128 # 6/5/4
        fdima2 = 64 # 3/2
        fdimb1 = 16 # 6/5/4/3
        fdimb2 = 12 # 2

        full=False
        self.f6 = butterfly4D(fdima1, fdimb1,withbn=withbn,full=full)
        self.p6 = sepConv4d(fdimb1,fdimb1, with_bn=False, full=full)

        self.f5 = butterfly4D(fdima1, fdimb1,withbn=withbn, full=full)
        self.p5 = sepConv4d(fdimb1,fdimb1, with_bn=False,full=full)

        self.f4 = butterfly4D(fdima1, fdimb1,withbn=withbn,full=full)
        self.p4 = sepConv4d(fdimb1,fdimb1, with_bn=False,full=full)

        self.f3 = butterfly4D(fdima2, fdimb1,withbn=withbn,full=full)
        self.p3 = sepConv4d(fdimb1,fdimb1, with_bn=False,full=full)

        ## soft WTA modules
        self.flow_reg64 = flow_reg(ent=use_entropy, maxdisp=self.md[0], fac=self.fac)
        self.flow_reg32 = flow_reg(ent=use_entropy, maxdisp=self.md[1])
        self.flow_reg16 = flow_reg(ent=use_entropy, maxdisp=self.md[2])
        self.flow_reg8 =  flow_reg(ent=use_entropy, maxdisp=self.md[3])

        ## warping modules
        self.warp5 = WarpModule()
        self.warp4 = WarpModule()
        self.warp3 = WarpModule()
        if self.training:
            self.warpx = WarpModule()
       

        ## hypotheses fusion modules, adopted from the refinement module of PWCNet
        # https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
        # c6
        self.dc6_conv1 = conv(128+4*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc6_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc6_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc6_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc6_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc6_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc6_conv7 = nn.Conv2d(32,2*fdimb1,kernel_size=3,stride=1,padding=1,bias=True)

        # c5
        self.dc5_conv1 = conv(128+4*fdimb1*2, 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc5_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc5_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc5_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc5_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc5_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc5_conv7 = nn.Conv2d(32,2*fdimb1*2,kernel_size=3,stride=1,padding=1,bias=True)

        # c4
        self.dc4_conv1 = conv(128+4*fdimb1*3, 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc4_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc4_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc4_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc4_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc4_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc4_conv7 = nn.Conv2d(32,2*fdimb1*3,kernel_size=3,stride=1,padding=1,bias=True)

        # c3
        self.dc3_conv1 = conv(64+16*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc3_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc3_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc3_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc3_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc3_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc3_conv7 = nn.Conv2d(32,8*fdimb1,kernel_size=3,stride=1,padding=1,bias=True)

        ## Out-of-range detection
        # if size[0]>1:  # only in train mode
        self.dc6_convo = nn.Sequential(conv(128+4*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1),
                            conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2),
                            conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4),
                            conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8),
                            conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16),
                            conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1),
                            nn.Conv2d(32,1,kernel_size=3,stride=1,padding=1,bias=True))

        self.dc5_convo = nn.Sequential(conv(128+2*4*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1),
                            conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2),
                            conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4),
                            conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8),
                            conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16),
                            conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1),
                            nn.Conv2d(32,1,kernel_size=3,stride=1,padding=1,bias=True))

        self.dc4_convo = nn.Sequential(conv(128+3*4*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1),
                            conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2),
                            conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4),
                            conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8),
                            conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16),
                            conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1),
                            nn.Conv2d(32,1,kernel_size=3,stride=1,padding=1,bias=True))

        self.dc3_convo = nn.Sequential(conv(64+16*fdimb1, 128, kernel_size=3, stride=1, padding=1,  dilation=1),
                            conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2),
                            conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4),
                            conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8),
                            conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16),
                            conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1),
                            nn.Conv2d(32,1,kernel_size=3,stride=1,padding=1,bias=True))

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                n = m.kernel_size[0] * m.kernel_size[1]*m.kernel_size[2] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if hasattr(m.bias,'data'):
                    m.bias.data.zero_()

    @staticmethod
    def add_model_specific_args(parent_parser=None):
        parent_parser = BaseModel.add_model_specific_args(parent_parser)
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--fac', type=float, default=1.0)
        parser.add_argument('--maxdisp', type=int, default=256)
        return parser
   
    #@profile
    def corr(self, refimg_fea, targetimg_fea, maxdisp, fac=1):
        """
        correlation function. Adopted from https://github.com/ClementPinard/Pytorch-Correlation-extension
        faster, but backwards not implemented.
        """
        from spatial_correlation_sampler import SpatialCorrelationSampler
        corr = SpatialCorrelationSampler(kernel_size=1,patch_size=(int(1+2*maxdisp//fac),int(1+2*maxdisp)),stride=1,padding=0,dilation_patch=1)
        cost = corr(refimg_fea, targetimg_fea)
        cost = F.leaky_relu(cost, 0.1,inplace=True)
        return cost


    def corrf(self, refimg_fea, targetimg_fea,maxdisp, fac=1):
        """
        another correlation function giving the same result as corr()
        supports backwards
        """
        b,c,height,width = refimg_fea.shape
        if refimg_fea.is_cuda:
            cost = Variable(torch.cuda.FloatTensor(b,c,2*maxdisp+1,2*int(maxdisp//fac)+1,height,width)).fill_(0.) # b,c,u,v,h,w
        else:
            cost = Variable(torch.FloatTensor(b,c,2*maxdisp+1,2*int(maxdisp//fac)+1,height,width)).fill_(0.) # b,c,u,v,h,w
        for i in range(2*maxdisp+1):
            ind = i-maxdisp
            for j in range(2*int(maxdisp//fac)+1):
                indd = j-int(maxdisp//fac)
                feata = refimg_fea[:,:,max(0,-indd):height-indd,max(0,-ind):width-ind]
                featb = targetimg_fea[:,:,max(0,+indd):height+indd,max(0,ind):width+ind]
                diff = (feata*featb)
                cost[:, :, i,j,max(0,-indd):height-indd,max(0,-ind):width-ind]   = diff
        cost = F.leaky_relu(cost, 0.1,inplace=True)
        return cost

    def get_oor_loss(self, flowl0, oor3, maxdisp, occ_mask):
        """ 
        return out-of-range loss
        """
        oor3_gt = (flowl0.abs() > maxdisp).detach() #  (8*self.md[3])
        oor3_gt = (((oor3_gt.sum(1)>0) + occ_mask)>0).float()  # oor, or occluded
        weights = oor3_gt.sum().float()/(oor3_gt.shape[0]*oor3_gt.shape[1]*oor3_gt.shape[2])
        weights = oor3_gt * (1-weights) + (1-oor3_gt) * weights
        loss_oor3 = F.binary_cross_entropy_with_logits(oor3,oor3_gt,size_average=True, weight=weights)
        return loss_oor3

    def weight_parameters(self):
        return [param for name, param in self.named_parameters() if 'weight' in name]

    def bias_parameters(self):
        return [param for name, param in self.named_parameters() if 'bias' in name]

    def forward(self, inputs, disc_aux=None):
        im = inputs['images']
        im = im.view(im.shape[0]*im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        bs = im.shape[0]//2

        c06,c05,c04,c03 = self.pspnet(im)
        c16 = c06[:bs];  c26 = c06[bs:]
        c15 = c05[:bs];  c25 = c05[bs:]
        c14 = c04[:bs];  c24 = c04[bs:]
        c13 = c03[:bs];  c23 = c03[bs:]

        # normalize
        c16n = c16 / (c16.norm(dim=1, keepdim=True)+1e-9)
        c26n = c26 / (c26.norm(dim=1, keepdim=True)+1e-9)
        c15n = c15 / (c15.norm(dim=1, keepdim=True)+1e-9)
        c25n = c25 / (c25.norm(dim=1, keepdim=True)+1e-9)
        c14n = c14 / (c14.norm(dim=1, keepdim=True)+1e-9)
        c24n = c24 / (c24.norm(dim=1, keepdim=True)+1e-9)
        c13n = c13 / (c13.norm(dim=1, keepdim=True)+1e-9)
        c23n = c23 / (c23.norm(dim=1, keepdim=True)+1e-9)

        up_flows = []
        oors = []
        flow_reg_maxs = []

        ## matching 6
        if self.training or (not im.is_cuda):
            feat6 = self.corrf(c16n,c26n,self.md[0],fac=self.fac)
        else:
            feat6 = self.corrf(c16n,c26n,self.md[0],fac=self.fac)
            #feat6 = self.corr(c16n,c26n,self.md[0],fac=self.fac)
        feat6 = self.f6(feat6) 
        cost6 = self.p6(feat6) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost6.shape
        cost6 = cost6.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow6h,ent6h = self.flow_reg64(cost6) # bx16, 2, h, w
        flow6h =  flow6h.view(bs,-1,h,w) # b, 16*2, h, w
        ent6h =  ent6h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg64.flowx.max())

        if self.training:
            x = torch.cat((ent6h.detach(), flow6h.detach(), c16),1)
            oor6 = self.dc6_convo(x)[:,0]
            oors.append(oor6)

        # hypotheses fusion
        x = torch.cat((ent6h.detach(), flow6h.detach(), c16),1)
        x = self.dc6_conv4(self.dc6_conv3(self.dc6_conv2(self.dc6_conv1(x))))
        va = self.dc6_conv7(self.dc6_conv6(self.dc6_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow6 = ( flow6h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)

        ## matching 5
        up_flow6 = F.interpolate(flow6, [im.size()[2]//32,im.size()[3]//32], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow6)
        warp5,_ = self.warp5(c25n, up_flow6)
        if self.training or (not im.is_cuda):
            feat5 = self.corrf(c15n,warp5,self.md[1])
        else:
            feat5 = self.corrf(c15n,warp5,self.md[1])
            #feat5 = self.corr(c15n,warp5,self.md[1])
        feat5 = self.f5(feat5) 
        cost5 = self.p5(feat5) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost5.shape
        cost5 = cost5.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow5h,ent5h = self.flow_reg32(cost5) # bx16, 2, h, w
        flow5h = flow5h.view(b,c,2,h,w) + up_flow6[:,np.newaxis]
        flow5h = flow5h.view(bs,-1,h,w) # b, 16*2, h, w
        ent5h =  ent5h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg32.flowx.max())

        # append coarse hypotheses
        flow5h = torch.cat((flow5h, F.interpolate(flow6h.detach()*2, [flow5h.shape[2],flow5h.shape[3]], mode='bilinear', align_corners=False)),1) # b, k2--k2, h, w
        ent5h = torch.cat((ent5h, F.interpolate(ent6h, [flow5h.shape[2],flow5h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent5h.detach(), flow5h.detach(), c15),1)
            oor5 = self.dc5_convo(x)[:,0]
            oors.append(oor5)

        # hypotheses fusion
        x = torch.cat((ent5h.detach(), flow5h.detach(), c15),1)
        x = self.dc5_conv4(self.dc5_conv3(self.dc5_conv2(self.dc5_conv1(x))))
        va5 = self.dc5_conv7(self.dc5_conv6(self.dc5_conv5(x)))
        va5 = va5.view(b,-1,2,h,w)
        flow5 = ( flow5h.view(b,-1,2,h,w) * F.softmax(va5,1) ).sum(1) # b, 2k, 2, h, w
        

        ## matching 4
        up_flow5 = F.interpolate(flow5, [im.size()[2]//16,im.size()[3]//16], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow5)
        warp4,_ = self.warp4(c24n, up_flow5)
        if self.training or (not im.is_cuda):
            feat4 = self.corrf(c14n,warp4,self.md[2])
        else:
            feat4 = self.corrf(c14n,warp4,self.md[2])
            #feat4 = self.corr(c14n,warp4,self.md[2])
        feat4 = self.f4(feat4) 
        cost4 = self.p4(feat4) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost4.shape
        cost4 = cost4.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow4h,ent4h = self.flow_reg16(cost4) # bx16, 2, h, w
        flow4h = flow4h.view(b,c,2,h,w) + up_flow5[:,np.newaxis]
        flow4h =  flow4h.view(bs,-1,h,w) # b, 16*2, h, w
        ent4h =  ent4h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg16.flowx.max())

        # append coarse hypotheses
        flow4h = torch.cat((flow4h, F.interpolate(flow5h.detach()*2, [flow4h.shape[2],flow4h.shape[3]], mode='bilinear', align_corners=False)),1)
        ent4h = torch.cat((ent4h, F.interpolate(ent5h, [flow4h.shape[2],flow4h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent4h.detach(), flow4h.detach(), c14),1)
            oor4 = self.dc4_convo(x)[:,0]
            oors.append(oor4)

        # hypotheses fusion
        x = torch.cat((ent4h.detach(), flow4h.detach(), c14),1)
        x =  self.dc4_conv4(self.dc4_conv3(self.dc4_conv2(self.dc4_conv1(x))))
        va = self.dc4_conv7(self.dc4_conv6(self.dc4_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow4 = ( flow4h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)


        ## matching 3
        up_flow4 = F.interpolate(flow4, [im.size()[2]//8,im.size()[3]//8], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow4)
        warp3,_ = self.warp3(c23n, up_flow4)
        if self.training or (not im.is_cuda):
            feat3 = self.corrf(c13n,warp3,self.md[3])
        else:
            feat3 = self.corrf(c13n,warp3,self.md[3])
            #feat3 = self.corr(c13n,warp3,self.md[3])
        feat3 = self.f3(feat3) 
        cost3 = self.p3(feat3) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost3.shape
        cost3 = cost3.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow3h,ent3h = self.flow_reg8(cost3) # bx16, 2, h, w
        flow3h = flow3h.view(b,c,2,h,w) + up_flow4[:,np.newaxis]
        flow3h = flow3h.view(bs,-1,h,w) # b, 16*2, h, w
        ent3h =  ent3h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg8.flowx.max())

        # append coarse hypotheses
        flow3h = torch.cat((flow3h, F.interpolate(flow4h.detach()*2, [flow3h.shape[2],flow3h.shape[3]], mode='bilinear', align_corners=False)),1)
        ent3h = torch.cat((ent3h, F.interpolate(ent4h, [flow3h.shape[2],flow3h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent3h.detach(), flow3h.detach(), c13),1)
            oor3 = self.dc3_convo(x)[:,0]
            oors.append(oor3)

        # hypotheses fusion
        x = torch.cat((ent3h.detach(), flow3h.detach(), c13),1)
        x = self.dc3_conv4(self.dc3_conv3(self.dc3_conv2(self.dc3_conv1(x))))
        va = self.dc3_conv7(self.dc3_conv6(self.dc3_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow3 = ( flow3h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)

        flow3 = F.interpolate(flow3, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow4 = F.interpolate(flow4, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow5 = F.interpolate(flow5, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow6 = F.interpolate(flow6, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow2 = flow3*2
        flows = [flow6, flow5, flow4, flow3, flow2]

        outputs = {}

        if self.training:
            outputs['flow_preds'] = flows
            outputs['up_flows_preds'] = up_flows
            outputs['oors_preds'] = oors
            outputs['flow_reg_maxs_preds'] = flow_reg_maxs
            outputs['flows'] = flow2[:, None]*4
        else:
            outputs['flows'] = flow2[:, None]*4

        return outputs


class VCN(VCNSmall):
    """
    VCN.
    md defines maximum displacement for each level, following a coarse-to-fine-warping scheme
    fac defines squeeze parameter for the coarsest level
    """

    pretrained_checkpoints = {
        'chairs': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-chairs-792544a0.ckpt',
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-things-afc0f455.ckpt',
        'sintel': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-sintel-5aac0540.ckpt',
        'kitti': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/vcn-kitti-29aa0c1a.ckpt'
    }

    def __init__(self, args, md=[4,4,4,4,4]):
        super(VCN, self).__init__(
            args=args, md=md)
        use_entropy = True
        withbn = True
        
        ## pspnet
        self.pspnet = pspnet(is_proj=False)

        ## Volumetric-UNet
        fdima1 = 128 # 6/5/4
        fdima2 = 64 # 3/2
        fdimb1 = 16 # 6/5/4/3
        fdimb2 = 12 # 2

        full=True
        self.f2 = butterfly4D(fdima2, fdimb2,withbn=withbn,full=full)
        self.p2 = sepConv4d(fdimb2,fdimb2, with_bn=False,full=full)

        ## soft WTA modules
        self.flow_reg4 =  flow_reg(ent=use_entropy, maxdisp=self.md[4])

        ## warping modules
        self.warp2 = WarpModule()
       

        ## hypotheses fusion modules, adopted from the refinement module of PWCNet
        # https://github.com/NVlabs/PWC-Net/blob/master/PyTorch/models/PWCNet.py
        # c2
        self.dc2_conv1 = conv(64+16*fdimb1+4*fdimb2, 128, kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc2_conv2 = conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2)
        self.dc2_conv3 = conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4)
        self.dc2_conv4 = conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8)
        self.dc2_conv5 = conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16)
        self.dc2_conv6 = conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1)
        self.dc2_conv7 = nn.Conv2d(32,4*2*fdimb1 + 2*fdimb2,kernel_size=3,stride=1,padding=1,bias=True)

        ## Out-of-range detection
        # if size[0]>1:  # only in train mode
        self.dc2_convo = nn.Sequential(conv(64+16*fdimb1+4*fdimb2, 128, kernel_size=3, stride=1, padding=1,  dilation=1),
                            conv(128,      128, kernel_size=3, stride=1, padding=2,  dilation=2),
                            conv(128,      128, kernel_size=3, stride=1, padding=4,  dilation=4),
                            conv(128,      96,  kernel_size=3, stride=1, padding=8,  dilation=8),
                            conv(96,       64,  kernel_size=3, stride=1, padding=16, dilation=16),
                            conv(64,       32,  kernel_size=3, stride=1, padding=1,  dilation=1),
                            nn.Conv2d(32,1,kernel_size=3,stride=1,padding=1,bias=True))

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                n = m.kernel_size[0] * m.kernel_size[1]*m.kernel_size[2] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if hasattr(m.bias,'data'):
                    m.bias.data.zero_()

    def forward(self, inputs, disc_aux=None):
        im = inputs['images']
        im = im.view(im.shape[0]*im.shape[1], im.shape[2], im.shape[3], im.shape[4])

        bs = im.shape[0]//2

        c06,c05,c04,c03,c02 = self.pspnet(im)
        c16 = c06[:bs];  c26 = c06[bs:]
        c15 = c05[:bs];  c25 = c05[bs:]
        c14 = c04[:bs];  c24 = c04[bs:]
        c13 = c03[:bs];  c23 = c03[bs:]
        c12 = c02[:bs];  c22 = c02[bs:]

        # normalize
        c16n = c16 / (c16.norm(dim=1, keepdim=True)+1e-9)
        c26n = c26 / (c26.norm(dim=1, keepdim=True)+1e-9)
        c15n = c15 / (c15.norm(dim=1, keepdim=True)+1e-9)
        c25n = c25 / (c25.norm(dim=1, keepdim=True)+1e-9)
        c14n = c14 / (c14.norm(dim=1, keepdim=True)+1e-9)
        c24n = c24 / (c24.norm(dim=1, keepdim=True)+1e-9)
        c13n = c13 / (c13.norm(dim=1, keepdim=True)+1e-9)
        c23n = c23 / (c23.norm(dim=1, keepdim=True)+1e-9)
        c12n = c12 / (c12.norm(dim=1, keepdim=True)+1e-9)
        c22n = c22 / (c22.norm(dim=1, keepdim=True)+1e-9)

        up_flows = []
        oors = []
        flow_reg_maxs = []

        ## matching 6
        if self.training or (not im.is_cuda):
            feat6 = self.corrf(c16n,c26n,self.md[0],fac=self.fac)
        else:
            feat6 = self.corrf(c16n,c26n,self.md[0],fac=self.fac)
            #feat6 = self.corr(c16n,c26n,self.md[0],fac=self.fac)
        feat6 = self.f6(feat6) 
        cost6 = self.p6(feat6) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost6.shape
        cost6 = cost6.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow6h,ent6h = self.flow_reg64(cost6) # bx16, 2, h, w
        flow6h =  flow6h.view(bs,-1,h,w) # b, 16*2, h, w
        ent6h =  ent6h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg64.flowx.max())

        if self.training:
            x = torch.cat((ent6h.detach(), flow6h.detach(), c16),1)
            oor6 = self.dc6_convo(x)[:,0]
            oors.append(oor6)

        # hypotheses fusion
        x = torch.cat((ent6h.detach(), flow6h.detach(), c16),1)
        x = self.dc6_conv4(self.dc6_conv3(self.dc6_conv2(self.dc6_conv1(x))))
        va = self.dc6_conv7(self.dc6_conv6(self.dc6_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow6 = ( flow6h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)

        ## matching 5
        up_flow6 = F.interpolate(flow6, [im.size()[2]//32,im.size()[3]//32], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow6)
        warp5,_ = self.warp5(c25n, up_flow6)
        if self.training or (not im.is_cuda):
            feat5 = self.corrf(c15n,warp5,self.md[1])
        else:
            feat5 = self.corrf(c15n,warp5,self.md[1])
            #feat5 = self.corr(c15n,warp5,self.md[1])
        feat5 = self.f5(feat5) 
        cost5 = self.p5(feat5) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost5.shape
        cost5 = cost5.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow5h,ent5h = self.flow_reg32(cost5) # bx16, 2, h, w
        flow5h = flow5h.view(b,c,2,h,w) + up_flow6[:,np.newaxis]
        flow5h = flow5h.view(bs,-1,h,w) # b, 16*2, h, w
        ent5h =  ent5h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg32.flowx.max())

        # append coarse hypotheses
        flow5h = torch.cat((flow5h, F.interpolate(flow6h.detach()*2, [flow5h.shape[2],flow5h.shape[3]], mode='bilinear', align_corners=False)),1) # b, k2--k2, h, w
        ent5h = torch.cat((ent5h, F.interpolate(ent6h, [flow5h.shape[2],flow5h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent5h.detach(), flow5h.detach(), c15),1)
            oor5 = self.dc5_convo(x)[:,0]
            oors.append(oor5)

        # hypotheses fusion
        x = torch.cat((ent5h.detach(), flow5h.detach(), c15),1)
        x = self.dc5_conv4(self.dc5_conv3(self.dc5_conv2(self.dc5_conv1(x))))
        va5 = self.dc5_conv7(self.dc5_conv6(self.dc5_conv5(x)))
        va5 = va5.view(b,-1,2,h,w)
        flow5 = ( flow5h.view(b,-1,2,h,w) * F.softmax(va5,1) ).sum(1) # b, 2k, 2, h, w
        

        ## matching 4
        up_flow5 = F.interpolate(flow5, [im.size()[2]//16,im.size()[3]//16], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow5)
        warp4,_ = self.warp4(c24n, up_flow5)
        if self.training or (not im.is_cuda):
            feat4 = self.corrf(c14n,warp4,self.md[2])
        else:
            feat4 = self.corrf(c14n,warp4,self.md[2])
            #feat4 = self.corr(c14n,warp4,self.md[2])
        feat4 = self.f4(feat4) 
        cost4 = self.p4(feat4) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost4.shape
        cost4 = cost4.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow4h,ent4h = self.flow_reg16(cost4) # bx16, 2, h, w
        flow4h = flow4h.view(b,c,2,h,w) + up_flow5[:,np.newaxis]
        flow4h =  flow4h.view(bs,-1,h,w) # b, 16*2, h, w
        ent4h =  ent4h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg16.flowx.max())

        # append coarse hypotheses
        flow4h = torch.cat((flow4h, F.interpolate(flow5h.detach()*2, [flow4h.shape[2],flow4h.shape[3]], mode='bilinear', align_corners=False)),1)
        ent4h = torch.cat((ent4h, F.interpolate(ent5h, [flow4h.shape[2],flow4h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent4h.detach(), flow4h.detach(), c14),1)
            oor4 = self.dc4_convo(x)[:,0]
            oors.append(oor4)

        # hypotheses fusion
        x = torch.cat((ent4h.detach(), flow4h.detach(), c14),1)
        x =  self.dc4_conv4(self.dc4_conv3(self.dc4_conv2(self.dc4_conv1(x))))
        va = self.dc4_conv7(self.dc4_conv6(self.dc4_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow4 = ( flow4h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)


        ## matching 3
        up_flow4 = F.interpolate(flow4, [im.size()[2]//8,im.size()[3]//8], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow4)
        warp3,_ = self.warp3(c23n, up_flow4)
        if self.training or (not im.is_cuda):
            feat3 = self.corrf(c13n,warp3,self.md[3])
        else:
            feat3 = self.corrf(c13n,warp3,self.md[3])
            #feat3 = self.corr(c13n,warp3,self.md[3])
        feat3 = self.f3(feat3) 
        cost3 = self.p3(feat3) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost3.shape
        cost3 = cost3.view(-1,u,v,h,w)  # bx16, 9,9,h,w, also predict uncertainty from here
        flow3h,ent3h = self.flow_reg8(cost3) # bx16, 2, h, w
        flow3h = flow3h.view(b,c,2,h,w) + up_flow4[:,np.newaxis]
        flow3h = flow3h.view(bs,-1,h,w) # b, 16*2, h, w
        ent3h =  ent3h.view(bs,-1,h,w) # b, 16*1, h, w
        flow_reg_maxs.append(self.flow_reg8.flowx.max())

        # append coarse hypotheses
        flow3h = torch.cat((flow3h, F.interpolate(flow4h.detach()*2, [flow3h.shape[2],flow3h.shape[3]], mode='bilinear', align_corners=False)),1)
        ent3h = torch.cat((ent3h, F.interpolate(ent4h, [flow3h.shape[2],flow3h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent3h.detach(), flow3h.detach(), c13),1)
            oor3 = self.dc3_convo(x)[:,0]
            oors.append(oor3)

        # hypotheses fusion
        x = torch.cat((ent3h.detach(), flow3h.detach(), c13),1)
        x = self.dc3_conv4(self.dc3_conv3(self.dc3_conv2(self.dc3_conv1(x))))
        va = self.dc3_conv7(self.dc3_conv6(self.dc3_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow3 = ( flow3h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)

        ## matching 2
        up_flow3 = F.interpolate(flow3, [im.size()[2]//4,im.size()[3]//4], mode='bilinear', align_corners=False)*2
        up_flows.append(up_flow3)
        warp2,_ = self.warp2(c22n, up_flow3)
        if self.training or (not im.is_cuda):
            feat2 = self.corrf(c12n,warp2,self.md[4])
        else:
            feat2 = self.corrf(c12n,warp2,self.md[4])
            #feat2 = self.corr(c12n,warp2,self.md[4])
        feat2 = self.f2(feat2) 
        cost2 = self.p2(feat2) # b, 16, u,v,h,w

        # soft WTA
        b,c,u,v,h,w = cost2.shape
        cost2 = cost2.view(-1,u,v,h,w)  # bx12, 9,9,h,w, also predict uncertainty from here
        flow2h,ent2h = self.flow_reg4(cost2) # bx12, 2, h, w
        flow2h = flow2h.view(b,c,2,h,w) + up_flow3[:,np.newaxis]
        flow2h = flow2h.view(bs,-1,h,w) # b, 12*2, h, w
        ent2h =  ent2h.view(bs,-1,h,w) # b, 12*1, h, w
        flow_reg_maxs.append(self.flow_reg4.flowx.max())

        # append coarse hypotheses
        flow2h = torch.cat((flow2h, F.interpolate(flow3h.detach()*2, [flow2h.shape[2],flow2h.shape[3]], mode='bilinear', align_corners=False)),1)
        ent2h = torch.cat((ent2h, F.interpolate(ent3h, [flow2h.shape[2],flow2h.shape[3]], mode='bilinear', align_corners=False)),1)

        if self.training:
            x = torch.cat((ent2h.detach(), flow2h.detach(), c12),1)
            oor2 = self.dc2_convo(x)[:,0]
            oors.append(oor2)

        # hypotheses fusion
        x = torch.cat((ent2h.detach(), flow2h.detach(), c12),1)
        x = self.dc2_conv4(self.dc2_conv3(self.dc2_conv2(self.dc2_conv1(x))))
        va = self.dc2_conv7(self.dc2_conv6(self.dc2_conv5(x)))
        va = va.view(b,-1,2,h,w)
        flow2 = ( flow2h.view(b,-1,2,h,w) * F.softmax(va,1) ).sum(1)


        flow2 = F.interpolate(flow2, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow3 = F.interpolate(flow3, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow4 = F.interpolate(flow4, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow5 = F.interpolate(flow5, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flow6 = F.interpolate(flow6, [im.size()[2],im.size()[3]], mode='bilinear', align_corners=False)
        flows = [flow6, flow5, flow4, flow3, flow2]

        outputs = {}

        if self.training:
            outputs['flow_preds'] = flows
            outputs['up_flows_preds'] = up_flows
            outputs['oors_preds'] = oors
            outputs['flow_reg_maxs_preds'] = flow_reg_maxs
            outputs['flows'] = flow2[:, None]*4
        else:
            outputs['flows'] = flow2[:, None]*4

        return outputs



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
import torch.nn.functional as tf

def conv(in_planes, out_planes, kernel_size=3, stride=1, dilation=1, isReLU=True):
    if isReLU:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, dilation=dilation,
                      padding=((kernel_size - 1) * dilation) // 2, bias=True),
            nn.LeakyReLU(0.1, inplace=True)
        )
    else:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, dilation=dilation,
                      padding=((kernel_size - 1) * dilation) // 2, bias=True)
        )


def upsample_factor2(inputs, target_as):
    inputs = tf.interpolate(inputs, scale_factor=2, mode="nearest")
    _, _, h, w = target_as.size()
    if inputs.size(2) != h or inputs.size(3) != w:
        return tf.interpolate(inputs, [h, w], mode="bilinear", align_corners=False)
    else:
        return inputs


class OccUpsampleNetwork(nn.Module):
    def __init__(self, ch_in, ch_out):
        super(OccUpsampleNetwork, self).__init__()

        self.feat_dim = 32
        self.init_conv = conv(ch_in, self.feat_dim)

        self.res_convs = nn.Sequential(
            conv(self.feat_dim, self.feat_dim),
            conv(self.feat_dim, self.feat_dim, isReLU=False)
        )
        self.res_end_conv = conv(self.feat_dim, self.feat_dim)
        self.mul_const = 0.1

        self.out_convs = conv(self.feat_dim, ch_out)

    def forward(self, occ, x):
        occ = upsample_factor2(occ, x)
        x_in = torch.cat([occ, x], dim=1)
        x_init = self.init_conv(x_in)
        x_res = x_init
        x_res = x_res + self.res_convs(x_res) * self.mul_const
        x_res = x_res + self.res_convs(x_res) * self.mul_const
        x_res = x_res + self.res_convs(x_res) * self.mul_const
        x_init = x_init + self.res_end_conv(x_res)

        return self.out_convs(x_init) + occ


def subtract_mean(input):
    return input - input.mean(2).mean(2).unsqueeze(2).unsqueeze(2).expand_as(input)

    
class RefineFlow(nn.Module):
    def __init__(self, ch_in):
        super(RefineFlow, self).__init__()

        self.kernel_size = 3
        self.pad_size = 1
        self.pad_ftn = nn.ReplicationPad2d(self.pad_size)

        self.convs = nn.Sequential(
            conv(ch_in, 128, 3, 1, 1),
            conv(128, 128, 3, 1, 1),
            conv(128, 64, 3, 1, 1),
            conv(64, 64, 3, 1, 1),
            conv(64, 32, 3, 1, 1),
            conv(32, 32, 3, 1, 1),
            conv(32, self.kernel_size * self.kernel_size, 3, 1, 1)
        )

        self.softmax_feat = nn.Softmax(dim=1)
        self.unfold_flow = nn.Unfold(kernel_size=(self.kernel_size, self.kernel_size))
        self.unfold_kernel = nn.Unfold(kernel_size=(1, 1))

    def forward(self, flow, diff_img, feature):
        b, _, h, w = flow.size()

        flow_m = subtract_mean(flow)
        norm2_img = torch.norm(diff_img, p=2, dim=1, keepdim=True)

        feat = self.convs(torch.cat([flow_m, norm2_img, feature], dim=1))
        feat_kernel = self.softmax_feat(-feat ** 2)

        flow_x = flow[:, 0].unsqueeze(1)
        flow_y = flow[:, 1].unsqueeze(1)

        flow_x_unfold = self.unfold_flow(self.pad_ftn(flow_x))
        flow_y_unfold = self.unfold_flow(self.pad_ftn(flow_y))
        feat_kernel_unfold = self.unfold_kernel(feat_kernel)

        flow_out_x = torch.sum(flow_x_unfold * feat_kernel_unfold, dim=1).unsqueeze(1).view(b, 1, h, w)
        flow_out_y = torch.sum(flow_y_unfold * feat_kernel_unfold, dim=1).unsqueeze(1).view(b, 1, h, w)

        return torch.cat([flow_out_x, flow_out_y], dim=1)


class RefineOcc(nn.Module):
    def __init__(self, ch_in):
        super(RefineOcc, self).__init__()

        self.kernel_size = 3
        self.pad_size = 1
        self.pad_ftn = nn.ReplicationPad2d(self.pad_size)

        self.convs = nn.Sequential(
            conv(ch_in, 128, 3, 1, 1),
            conv(128, 128, 3, 1, 1),
            conv(128, 64, 3, 1, 1),
            conv(64, 64, 3, 1, 1),
            conv(64, 32, 3, 1, 1),
            conv(32, 32, 3, 1, 1),
            conv(32, self.kernel_size * self.kernel_size, 3, 1, 1)
        )

        self.softmax_feat = nn.Softmax(dim=1)
        self.unfold_occ = nn.Unfold(kernel_size=(self.kernel_size, self.kernel_size))
        self.unfold_kernel = nn.Unfold(kernel_size=(1, 1))

    def forward(self, occ, feat1, feat2):
        b, _, h, w = occ.size()

        feat = self.convs(torch.cat([occ, feat1, feat2], dim=1))
        feat_kernel = self.softmax_feat(-feat ** 2)

        occ_unfold = self.unfold_occ(self.pad_ftn(occ))
        feat_kernel_unfold = self.unfold_kernel(feat_kernel)

        occ_out = torch.sum(occ_unfold * feat_kernel_unfold, dim=1).unsqueeze(1).view(b, 1, h, w)

        return occ_out


'''
Portions of this code copyright 2017, Clement Pinard
'''
from argparse import Namespace

import torch
import torch.nn as nn
from torch.nn import init

from .submodules import *
from .flownet_base import FlowNetBase


class FlowNetFusion(FlowNetBase):
    def __init__(self,
                 args: Namespace):
        args.loss_start_scale = 1
        args.loss_num_scales = 3
        super(FlowNetFusion, self).__init__(args)

        self.conv0   = conv(self.args.batch_norm,  11,   64)
        self.conv1   = conv(self.args.batch_norm,  64,   64, stride=2)
        self.conv1_1 = conv(self.args.batch_norm,  64,   128)
        self.conv2   = conv(self.args.batch_norm,  128,  128, stride=2)
        self.conv2_1 = conv(self.args.batch_norm,  128,  128)

        self.deconv1 = deconv(128,32)
        self.deconv0 = deconv(162,16)

        self.inter_conv1 = i_conv(self.args.batch_norm,  162,   32)
        self.inter_conv0 = i_conv(self.args.batch_norm,  82,   16)

        self.predict_flow2 = predict_flow(128)
        self.predict_flow1 = predict_flow(32)
        self.predict_flow0 = predict_flow(16)

        self.upsampled_flow2_to_1 = nn.ConvTranspose2d(2, 2, 4, 2, 1)
        self.upsampled_flow1_to_0 = nn.ConvTranspose2d(2, 2, 4, 2, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

    def forward(self, inputs):
        x = inputs['images']

        x = x.view(x.shape[0], x.shape[1]*x.shape[2], x.shape[3], x.shape[4])
        out_conv0 = self.conv0(x)
        out_conv1 = self.conv1_1(self.conv1(out_conv0))
        out_conv2 = self.conv2_1(self.conv2(out_conv1))

        flow2       = self.predict_flow2(out_conv2)
        flow2_up    = self.upsampled_flow2_to_1(flow2)
        out_deconv1 = self.deconv1(out_conv2)
        
        concat1 = torch.cat((out_conv1,out_deconv1,flow2_up),1)
        out_interconv1 = self.inter_conv1(concat1)
        flow1       = self.predict_flow1(out_interconv1)
        flow1_up    = self.upsampled_flow1_to_0(flow1)
        out_deconv0 = self.deconv0(concat1)
        
        concat0 = torch.cat((out_conv0,out_deconv0,flow1_up),1)
        out_interconv0 = self.inter_conv0(concat0)
        flow0       = self.predict_flow0(out_interconv0)

        outputs = {}

        if self.training:
            outputs['flow_preds'] = [flow0.float(),flow1.float(),flow2.float()]
            outputs['flows'] = flow0[:, None]
        else:
            outputs['flows'] = flow0[:, None]

        return outputs



"""Provide useful functions for using PTLFlow."""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

__version__ = '0.2.0'

import logging
from argparse import Namespace
from pathlib import Path
from typing import List, Optional

import requests
import torch
from torch import hub

from ptlflow.models.base_model.base_model import BaseModel
from ptlflow.models.flownet.flownet2 import FlowNet2
from ptlflow.models.flownet.flownetc import FlowNetC
from ptlflow.models.flownet.flownetcs import FlowNetCS
from ptlflow.models.flownet.flownetcss import FlowNetCSS
from ptlflow.models.flownet.flownets import FlowNetS
from ptlflow.models.flownet.flownetsd import FlowNetSD
from ptlflow.models.hd3.hd3 import HD3, HD3Context
from ptlflow.models.irr.pwcnet import IRRPWCNet
from ptlflow.models.irr.pwcnet_irr import IRRPWCNetIRR
from ptlflow.models.irr.irr_pwc import IRRPWC
from ptlflow.models.liteflownet.liteflownet import LiteFlowNet
from ptlflow.models.liteflownet.liteflownet3 import (
    LiteFlowNet3, LiteFlowNet3PseudoReg, LiteFlowNet3S, LiteFlowNet3SPseudoReg)
from ptlflow.models.liteflownet.liteflownet2 import LiteFlowNet2, LiteFlowNet2PseudoReg
from ptlflow.models.pwcnet.pwcnet import PWCNet, PWCDCNet
from ptlflow.models.raft.raft import RAFT, RAFTSmall
from ptlflow.models.scopeflow.irr_pwc_v2 import ScopeFlow
from ptlflow.models.vcn.vcn import VCN, VCNSmall
from ptlflow.utils.utils import config_logging

config_logging()


models_dict = {
    'flownet2': FlowNet2,
    'flownetc': FlowNetC,
    'flownetcs': FlowNetCS,
    'flownetcss': FlowNetCSS,
    'flownets': FlowNetS,
    'flownetsd': FlowNetSD,
    'hd3': HD3,
    'hd3_ctxt': HD3Context,
    'irr_pwc': IRRPWC,
    'irr_pwcnet': IRRPWCNet,
    'irr_pwcnet_irr': IRRPWCNetIRR,
    'liteflownet': LiteFlowNet,
    'liteflownet2': LiteFlowNet2,
    'liteflownet2_pseudoreg': LiteFlowNet2PseudoReg,
    'liteflownet3': LiteFlowNet3,
    'liteflownet3_pseudoreg': LiteFlowNet3PseudoReg,
    'liteflownet3s': LiteFlowNet3S,
    'liteflownet3s_pseudoreg': LiteFlowNet3SPseudoReg,
    'pwcnet': PWCNet,
    'pwcdcnet': PWCDCNet,
    'raft': RAFT,
    'raft_small': RAFTSmall,
    'scopeflow': ScopeFlow,
    'vcn': VCN,
    'vcn_small': VCNSmall,
}


def download_scripts(
    destination_dir: Path = Path('ptlflow_scripts')
) -> None:
    """Download the main scripts and configs to start working with PTLFlow."""
    github_url = 'https://raw.githubusercontent.com/hmorimitsu/ptlflow/main/'
    script_names = [
        'datasets.yml',
        'infer.py',
        'train.py',
        'validate.py'
    ]

    destination_dir.mkdir(parents=True, exist_ok=True)

    for sname in script_names:
        script_url = github_url + sname
        data = requests.get(script_url)
        if data.status_code == 200:
            with open(destination_dir / sname, 'wb') as f:
                f.write(data.content)
        else:
            logging.warning('Script %s was not found.', script_url)

    logging.info('Downloaded scripts to %s.', str(destination_dir))


def get_model(
    model_name: str,
    pretrained_ckpt: Optional[str] = None,
    args: Optional[Namespace] = None
) -> BaseModel:
    """Return an instance of a chosen model.

    The instance can have configured by he arguments, and load some existing pretrained weights.

    Note that this is different from get_model_reference(), which returns a reference to the model class. The instance,
    returned by this function, is a class already instantiated. Therefore, the return of this function is equivalent to
    "return get_model_reference()()", which looks confusing. This can be rewritten as
    "model_ref = get_model_reference(); return model_ref()".

    Parameters
    ----------
    model_name : str
        Name of the model to get an instance of.
    pretrained_ckpt : Optional[str], optional
        Name of the pretrained weight to load or a path to a local checkpoint file.
    args : Optional[Namespace], optional
        Some arguments that ill be provided to the model.

    Returns
    -------
    BaseModel
        The instance of the chosen model.

    Raises
    ------
    ValueError
        If the given checkpoint name is not a valid choice.
    ValueError
        If a checkpoint name is given, but the model does not have any pretrained weights available.

    See Also
    --------
    get_model_reference : To get a reference to the class of a model.
    """
    model_ref = get_model_reference(model_name)
    if args is None:
        parser = model_ref.add_model_specific_args()
        args = parser.parse_args([])
    model = model_ref(args)

    if pretrained_ckpt is None and args is not None and args.pretrained_ckpt is not None:
        pretrained_ckpt = args.pretrained_ckpt

    if pretrained_ckpt is not None:
        if Path(pretrained_ckpt).exists():
            ckpt_path = pretrained_ckpt
        elif hasattr(model_ref, 'pretrained_checkpoints'):
            ckpt_path = model_ref.pretrained_checkpoints.get(pretrained_ckpt)
            if ckpt_path is None:
                raise ValueError(
                    f'Invalid checkpoint name {pretrained_ckpt}. '
                    f'Choose one from {{{",".join(model.pretrained_checkpoints.keys())}}}')
        else:
            raise ValueError(f'Cannot find checkpoint {pretrained_ckpt} for model {model_name}')

        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        if Path(ckpt_path).exists():
            ckpt = torch.load(ckpt_path, map_location=torch.device(device))
        else:
            model_dir = Path(hub.get_dir()) / 'ptlflow' / 'checkpoints'
            ckpt = hub.load_state_dict_from_url(
                ckpt_path, model_dir=model_dir, map_location=torch.device(device), check_hash=True)

        state_dict = ckpt['state_dict']
        model.load_state_dict(state_dict)
    return model


def get_model_reference(
    model_name: str
) -> BaseModel:
    """Return a reference to the class of a chosen model.

    Note that this is different from get_model(), which returns an instance of a model. The reference, returned by this
    function, is a class before instantiation. Therefore, the return of this function can be used to instantiate a model as
    "model_ref = get_model_reference(); model_instance = model_ref()".

    Parameters
    ----------
    model_name : str
        Name of the model to get a reference of.

    Returns
    -------
    BaseModel
        A reference to the chosen model.

    Raises
    ------
    ValueError
        If the given name is not a valid choice.

    See Also
    --------
    get_model : To get an instance of a model.
    """
    try:
        return models_dict[model_name]
    except KeyError:
        raise ValueError(f'Unknown model name: {model_name}. Choose from [{", ".join(models_dict.keys())}]')


def get_trainable_model_names() -> List[str]:
    """Return a list of model names that are able to be trained.
    
    This function return the names of the model that have a loss function defined.

    Returns
    =======
    List[str]
        The list of the model names that can be trained.
    """
    return [mname for mname in models_dict.keys() if get_model(mname).loss_fn is not None]



import pdb
import torch.nn as nn
import math
import torch
from torch.nn.parameter import Parameter
import torch.nn.functional as F
from torch.nn import Module
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _quadruple
from torch.autograd import Variable
from torch.nn import Conv2d

def conv4d(data,filters,bias=None,permute_filters=True,use_half=False):
    """
    This is done by stacking results of multiple 3D convolutions, and is very slow.
    Taken from https://github.com/ignacio-rocco/ncnet
    """
    b,c,h,w,d,t=data.size()

    data=data.permute(2,0,1,3,4,5).contiguous() # permute to avoid making contiguous inside loop    
        
    # Same permutation is done with filters, unless already provided with permutation
    if permute_filters:
        filters=filters.permute(2,0,1,3,4,5).contiguous() # permute to avoid making contiguous inside loop    

    c_out=filters.size(1)
    if use_half:
        output = Variable(torch.HalfTensor(h,b,c_out,w,d,t),requires_grad=data.requires_grad)
    else:
        output = Variable(torch.zeros(h,b,c_out,w,d,t),requires_grad=data.requires_grad)
    
    padding=filters.size(0)//2
    if use_half:
        Z=Variable(torch.zeros(padding,b,c,w,d,t).half())
    else:
        Z=Variable(torch.zeros(padding,b,c,w,d,t))
    
    if data.is_cuda:
        Z=Z.cuda(data.get_device())    
        output=output.cuda(data.get_device())
        
    data_padded = torch.cat((Z,data,Z),0)
    

    for i in range(output.size(0)): # loop on first feature dimension
        # convolve with center channel of filter (at position=padding)
        output[i,:,:,:,:,:]=F.conv3d(data_padded[i+padding,:,:,:,:,:], 
                                     filters[padding,:,:,:,:,:], bias=bias, stride=1, padding=padding)
        # convolve with upper/lower channels of filter (at postions [:padding] [padding+1:])
        for p in range(1,padding+1):
            output[i,:,:,:,:,:]=output[i,:,:,:,:,:]+F.conv3d(data_padded[i+padding-p,:,:,:,:,:], 
                                                             filters[padding-p,:,:,:,:,:], bias=None, stride=1, padding=padding)
            output[i,:,:,:,:,:]=output[i,:,:,:,:,:]+F.conv3d(data_padded[i+padding+p,:,:,:,:,:], 
                                                             filters[padding+p,:,:,:,:,:], bias=None, stride=1, padding=padding)

    output=output.permute(1,2,0,3,4,5).contiguous()
    return output

class Conv4d(_ConvNd):
    """Applies a 4D convolution over an input signal composed of several input
    planes.
    """

    def __init__(self, in_channels, out_channels, kernel_size, bias=True, pre_permuted_filters=True): 
        # stride, dilation and groups !=1 functionality not tested 
        stride=1
        dilation=1
        groups=1
        # zero padding is added automatically in conv4d function to preserve tensor size
        padding = 0
        kernel_size = _quadruple(kernel_size)
        stride = _quadruple(stride)
        padding = _quadruple(padding)
        dilation = _quadruple(dilation)
        super(Conv4d, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _quadruple(0), groups, bias)  
        # weights will be sliced along one dimension during convolution loop
        # make the looping dimension to be the first one in the tensor, 
        # so that we don't need to call contiguous() inside the loop
        self.pre_permuted_filters=pre_permuted_filters
        if self.pre_permuted_filters:
            self.weight.data=self.weight.data.permute(2,0,1,3,4,5).contiguous()
        self.use_half=False
    #    self.isbias = bias
    #    if not self.isbias:
    #        self.bn = torch.nn.BatchNorm1d(out_channels)


    def forward(self, input):
        out = conv4d(input, self.weight, bias=self.bias,permute_filters=not self.pre_permuted_filters,use_half=self.use_half) # filters pre-permuted in constructor
    #    if not self.isbias:
    #        b,c,u,v,h,w = out.shape
    #        out = self.bn(out.view(b,c,-1)).view(b,c,u,v,h,w)
        return out

class fullConv4d(torch.nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, bias=True, pre_permuted_filters=True):
        super(fullConv4d, self).__init__()
        self.conv = Conv4d(in_channels, out_channels, kernel_size, bias=bias, pre_permuted_filters=pre_permuted_filters)
        self.isbias = bias
        if not self.isbias:
            self.bn = torch.nn.BatchNorm1d(out_channels)

    def forward(self, input):
        out = self.conv(input)
        if not self.isbias:
            b,c,u,v,h,w = out.shape
            out = self.bn(out.view(b,c,-1)).view(b,c,u,v,h,w)
        return out

class butterfly4D(torch.nn.Module):
    '''
    butterfly 4d
    '''
    def __init__(self, fdima, fdimb, withbn=True, full=True,groups=1):
        super(butterfly4D, self).__init__()
        self.proj = nn.Sequential(projfeat4d(fdima, fdimb, 1, with_bn=withbn,groups=groups),
                                  nn.ReLU(inplace=True),)
        self.conva1 = sepConv4dBlock(fdimb,fdimb,with_bn=withbn, stride=(2,1,1),full=full,groups=groups)
        self.conva2 = sepConv4dBlock(fdimb,fdimb,with_bn=withbn, stride=(2,1,1),full=full,groups=groups)
        self.convb3 = sepConv4dBlock(fdimb,fdimb,with_bn=withbn, stride=(1,1,1),full=full,groups=groups)
        self.convb2 = sepConv4dBlock(fdimb,fdimb,with_bn=withbn, stride=(1,1,1),full=full,groups=groups)
        self.convb1 = sepConv4dBlock(fdimb,fdimb,with_bn=withbn, stride=(1,1,1),full=full,groups=groups)

    #@profile
    def forward(self,x):
        out = self.proj(x)
        b,c,u,v,h,w = out.shape # 9x9

        out1 = self.conva1(out) # 5x5, 3
        _,c1,u1,v1,h1,w1 = out1.shape

        out2 = self.conva2(out1) # 3x3, 9
        _,c2,u2,v2,h2,w2 = out2.shape

        out2 = self.convb3(out2) # 3x3, 9

        tout1 = F.interpolate(out2.view(b,c,u2,v2,-1),(u1,v1,h2*w2),mode='trilinear', align_corners=False).view(b,c,u1,v1,h2,w2) # 5x5
        tout1 = F.interpolate(tout1.view(b,c,-1,h2,w2),(u1*v1,h1,w1),mode='trilinear', align_corners=False).view(b,c,u1,v1,h1,w1) # 5x5
        out1 = tout1 + out1
        out1 = self.convb2(out1)

        tout = F.interpolate(out1.view(b,c,u1,v1,-1),(u,v,h1*w1),mode='trilinear', align_corners=False).view(b,c,u,v,h1,w1)
        tout = F.interpolate(tout.view(b,c,-1,h1,w1),(u*v,h,w),mode='trilinear', align_corners=False).view(b,c,u,v,h,w)
        out = tout + out
        out = self.convb1(out)

        return out



class projfeat4d(torch.nn.Module):
    '''
    Turn 3d projection into 2d projection
    '''
    def __init__(self, in_planes, out_planes, stride, with_bn=True,groups=1):
        super(projfeat4d, self).__init__()
        self.with_bn = with_bn
        self.stride = stride
        self.conv1 = nn.Conv3d(in_planes, out_planes, 1, (stride,stride,1), padding=0,bias=not with_bn,groups=groups)
        self.bn = nn.BatchNorm3d(out_planes)

    def forward(self,x):
        b,c,u,v,h,w = x.size()
        x = self.conv1(x.view(b,c,u,v,h*w))
        if self.with_bn:
            x = self.bn(x)
        _,c,u,v,_ = x.shape
        x = x.view(b,c,u,v,h,w)
        return x

class sepConv4d(torch.nn.Module):
    '''
    Separable 4d convolution block as 2 3D convolutions
    '''
    def __init__(self, in_planes, out_planes, stride=(1,1,1), with_bn=True, ksize=3, full=True,groups=1):
        super(sepConv4d, self).__init__()
        bias = not with_bn
        self.isproj = False
        self.stride = stride[0]
        expand = 1

        if with_bn:
            if in_planes != out_planes:
                self.isproj = True
                self.proj = nn.Sequential(nn.Conv2d(in_planes, out_planes, 1, bias=bias, padding=0,groups=groups),
                                          nn.BatchNorm2d(out_planes))
            if full:
                self.conv1 = nn.Sequential(nn.Conv3d(in_planes*expand, in_planes, (1,ksize,ksize), stride=(1,self.stride,self.stride), bias=bias, padding=(0,ksize//2,ksize//2),groups=groups),
                                           nn.BatchNorm3d(in_planes))
            else:
                self.conv1 = nn.Sequential(nn.Conv3d(in_planes*expand, in_planes, (1,ksize,ksize), stride=1,                           bias=bias, padding=(0,ksize//2,ksize//2),groups=groups),
                                           nn.BatchNorm3d(in_planes))
            self.conv2 = nn.Sequential(nn.Conv3d(in_planes, in_planes*expand, (ksize,ksize,1), stride=(self.stride,self.stride,1), bias=bias, padding=(ksize//2,ksize//2,0),groups=groups),
                                       nn.BatchNorm3d(in_planes*expand))
        else:
            if in_planes != out_planes:
                self.isproj = True
                self.proj = nn.Conv2d(in_planes, out_planes, 1, bias=bias, padding=0,groups=groups)
            if full:
                self.conv1 = nn.Conv3d(in_planes*expand, in_planes, (1,ksize,ksize), stride=(1,self.stride,self.stride), bias=bias, padding=(0,ksize//2,ksize//2),groups=groups)
            else:
                self.conv1 = nn.Conv3d(in_planes*expand, in_planes, (1,ksize,ksize), stride=1,                           bias=bias, padding=(0,ksize//2,ksize//2),groups=groups)
            self.conv2 = nn.Conv3d(in_planes, in_planes*expand, (ksize,ksize,1), stride=(self.stride,self.stride,1), bias=bias, padding=(ksize//2,ksize//2,0),groups=groups)
        self.relu = nn.ReLU(inplace=True)
        
    #@profile
    def forward(self,x):
        b,c,u,v,h,w = x.shape
        x = self.conv2(x.view(b,c,u,v,-1)) # WTA convolution over (u,v)
        b,c,u,v,_ = x.shape
        x = self.relu(x)
        x = self.conv1(x.view(b,c,-1,h,w)) # spatial convolution over (x,y)
        b,c,_,h,w = x.shape

        if self.isproj:
            x = self.proj(x.view(b,c,-1,w))
        x = x.view(b,-1,u,v,h,w)
        return x


class sepConv4dBlock(torch.nn.Module):
    '''
    Separable 4d convolution block as 2 2D convolutions and a projection
    layer
    '''
    def __init__(self, in_planes, out_planes, stride=(1,1,1), with_bn=True, full=True,groups=1):
        super(sepConv4dBlock, self).__init__()
        if in_planes == out_planes and stride==(1,1,1):
            self.downsample = None
        else:
            if full:
                self.downsample = sepConv4d(in_planes, out_planes, stride, with_bn=with_bn,ksize=1, full=full,groups=groups)
            else:
                self.downsample = projfeat4d(in_planes, out_planes,stride[0], with_bn=with_bn,groups=groups)
        self.conv1 = sepConv4d(in_planes, out_planes, stride, with_bn=with_bn, full=full ,groups=groups)
        self.conv2 = sepConv4d(out_planes, out_planes,(1,1,1), with_bn=with_bn, full=full,groups=groups)
        self.relu1 = nn.ReLU(inplace=True)
        self.relu2 = nn.ReLU(inplace=True)

    #@profile
    def forward(self,x):
        out = self.relu1(self.conv1(x))
        if self.downsample:
            x = self.downsample(x)
        out = self.relu2(x + self.conv2(out))
        return out


##import torch.backends.cudnn as cudnn
##cudnn.benchmark = True
#import time
##im = torch.randn(9,64,9,160,224).cuda()
##net = torch.nn.Conv3d(64, 64, 3).cuda()
##net = Conv4d(1,1,3,bias=True,pre_permuted_filters=True).cuda()
##net = sepConv4dBlock(2,2,stride=(1,1,1)).cuda()
#
##im = torch.randn(1,16,9,9,96,320).cuda()
##net = sepConv4d(16,16,with_bn=False).cuda()
#
##im = torch.randn(1,16,81,96,320).cuda()
##net = torch.nn.Conv3d(16,16,(1,3,3),padding=(0,1,1)).cuda()
#
##im = torch.randn(1,16,9,9,96*320).cuda()
##net = torch.nn.Conv3d(16,16,(3,3,1),padding=(1,1,0)).cuda()
#
##im = torch.randn(10000,10,9,9).cuda()
##net = torch.nn.Conv2d(10,10,3,padding=1).cuda()
#
##im = torch.randn(81,16,96,320).cuda()
##net = torch.nn.Conv2d(16,16,3,padding=1).cuda()
#c=   int(16 *1)
#cp = int(16 *1)
#h=int(96  *4)
#w=int(320 *4)
#k=3
#im = torch.randn(1,c,h,w).cuda()
#net = torch.nn.Conv2d(c,cp,k,padding=k//2).cuda()
#
#im2 = torch.randn(cp,k*k*c).cuda()
#im1 = F.unfold(im, (k,k), padding=k//2)[0]
# 
#
#net(im)
#net(im)
#torch.mm(im2,im1)
#torch.mm(im2,im1)
#torch.cuda.synchronize()
#beg = time.time()
#for i in range(100):
#    net(im)
#    #im1 = F.unfold(im, (k,k), padding=k//2)[0]
#    torch.mm(im2,im1)
#torch.cuda.synchronize()
#print('%f'%((time.time()-beg)*10.))


import torch
import torch.nn as nn
import torch.nn.functional as F
from .hd3_ops import *


class LossCalculator(object):

    def __init__(self, task):
        assert task in ['flow', 'stereo']
        self.task = task
        self.dim = 1 if task == 'stereo' else 2

    def __call__(self, outputs, inputs):
        ms_prob = outputs['ms_prob']
        ms_pred = outputs['ms_pred']
        corr_range = outputs['corr_range']
        ds = outputs['downsample']
        gt = inputs['flows'][:, 0]
        B, C, H, W = gt.size()
        lv = len(ms_prob)
        criterion = nn.KLDivLoss(reduction='batchmean').to(device=gt.device)
        losses = {}
        kld_loss = 0
        for l in range(lv):
            scaled_gt, valid_mask = downsample_flow(gt, 1 / 2**(ds - l))
            if self.task == 'stereo':
                scaled_gt = scaled_gt[:, 0, :, :].unsqueeze(1)
            if l > 0:
                scaled_gt = scaled_gt - F.interpolate(
                    ms_pred[l - 1],
                    scale_factor=2,
                    mode='bilinear',
                    align_corners=True)
            scaled_gt = scaled_gt / 2**(ds - l)
            gt_dist = vector2density(scaled_gt, corr_range[l],
                                     self.dim) * valid_mask
            kld_loss += 4**(ds - l) / (H * W) * criterion(
                F.log_softmax(ms_prob[l], dim=1), gt_dist.detach())

        losses['loss'] = kld_loss
        for loss_type, loss_value in losses.items():
            losses[loss_type] = loss_value.reshape(1)
        return losses


def EndPointError(output, gt):
    # output: [B, 1/2, H, W], stereo or flow prediction
    # gt: [B, C, H, W], 2D ground-truth annotation which may contain a mask
    # NOTE: To benchmark the result, please ensure the ground-truth keeps
    # its ORIGINAL RESOLUTION.
    if output.size(1) == 1:  # stereo
        output = disp2flow(output)
    output = resize_dense_vector(output, gt.size(2), gt.size(3))
    error = torch.norm(output - gt[:, :2, :, :], 2, 1, keepdim=False)
    if gt.size(1) == 3:
        mask = (gt[:, 2, :, :] > 0).float()
    else:
        mask = torch.ones_like(error)
    epe = (error * mask).sum() / mask.sum()
    return epe.reshape(1)



import torch
import torch.nn as nn
import torch.nn.functional as F


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn='group', stride=1):
        super(ResidualBlock, self).__init__()
  
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, padding=1, stride=stride)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == 'group':
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
        
        elif norm_fn == 'batch':
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)
        
        elif norm_fn == 'instance':
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == 'none':
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None
        
        else:    
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3)


    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x+y)



class BottleneckBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn='group', stride=1):
        super(BottleneckBlock, self).__init__()
  
        self.conv1 = nn.Conv2d(in_planes, planes//4, kernel_size=1, padding=0)
        self.conv2 = nn.Conv2d(planes//4, planes//4, kernel_size=3, padding=1, stride=stride)
        self.conv3 = nn.Conv2d(planes//4, planes, kernel_size=1, padding=0)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == 'group':
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes//4)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes//4)
            self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm4 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
        
        elif norm_fn == 'batch':
            self.norm1 = nn.BatchNorm2d(planes//4)
            self.norm2 = nn.BatchNorm2d(planes//4)
            self.norm3 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.BatchNorm2d(planes)
        
        elif norm_fn == 'instance':
            self.norm1 = nn.InstanceNorm2d(planes//4)
            self.norm2 = nn.InstanceNorm2d(planes//4)
            self.norm3 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.InstanceNorm2d(planes)

        elif norm_fn == 'none':
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()
            if not stride == 1:
                self.norm4 = nn.Sequential()

        if stride == 1:
            self.downsample = None
        
        else:    
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm4)


    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))
        y = self.relu(self.norm3(self.conv3(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x+y)

class BasicEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn='batch', dropout=0.0):
        super(BasicEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == 'group':
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)
            
        elif self.norm_fn == 'batch':
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == 'instance':
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == 'none':
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64,  stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)
        
        self.in_planes = dim
        return nn.Sequential(*layers)


    def forward(self, x):

        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


class SmallEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn='batch', dropout=0.0):
        super(SmallEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == 'group':
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=32)
            
        elif self.norm_fn == 'batch':
            self.norm1 = nn.BatchNorm2d(32)

        elif self.norm_fn == 'instance':
            self.norm1 = nn.InstanceNorm2d(32)

        elif self.norm_fn == 'none':
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 32, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 32
        self.layer1 = self._make_layer(32,  stride=1)
        self.layer2 = self._make_layer(64, stride=2)
        self.layer3 = self._make_layer(96, stride=2)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)
        
        self.conv2 = nn.Conv2d(96, output_dim, kernel_size=1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = BottleneckBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = BottleneckBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)
    
        self.in_planes = dim
        return nn.Sequential(*layers)


    def forward(self, x):

        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim, batch_dim], dim=0)

        return x


"""Functions taken and adapted from https://github.com/princeton-vl/RAFT/blob/master/core/utils/utils.py."""

# Original license below

# BSD 3-Clause License

# Copyright (c) 2020, princeton-vl
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

import numpy as np
import torch.nn.functional as F


class InputPadder:
    """Pads images such that dimensions are divisible by stride."""
    def __init__(self, dims, stride=8):
        self.ht, self.wd = dims[-2:]
        pad_ht = (((self.ht // stride) + 1) * stride - self.ht) % stride
        pad_wd = (((self.wd // stride) + 1) * stride - self.wd) % stride
        if len(dims) < 5:
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, 0, pad_ht]
        else:
            self._pad = [pad_wd//2, pad_wd - pad_wd//2, 0, pad_ht, 0, 0]

    def pad(self, x):
        return F.pad(x, self._pad, mode='replicate')

    def unpad(self, x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht-self._pad[3], self._pad[0], wd-self._pad[1]]
        return x[..., c[0]:c[1], c[2]:c[3]]


def read_pfm(
    file_path: str
) -> np.ndarray:
    """Read optical flow files in PFM format.

    Taken and adapted from https://github.com/princeton-vl/RAFT/blob/master/core/utils/frame_utils.py.

    Parameters
    ----------
    file_path : str
        Path to the optical flow file.

    Returns
    -------
    np.ndarray
        The optical flow in HC format.

    Raises
    ------
    ValueError
        If the file is not in valid PFM format.
    ValueError
        If the file header is corrupted.
    """
    with open(file_path, 'rb') as f:
        color = None
        width = None
        height = None
        scale = None
        endian = None

        header = f.readline().rstrip()
        if header == b'PF':
            color = True
        elif header == b'Pf':
            color = False
        else:
            raise ValueError('Not a PFM file.')

        dim_match = re.match(rb'^(\d+)\s(\d+)\s$', f.readline())
        if dim_match:
            width, height = map(int, dim_match.groups())
        else:
            raise ValueError('Malformed PFM header.')

        scale = float(f.readline().rstrip())
        if scale < 0:  # little-endian
            endian = '<'
            scale = -scale
        else:
            endian = '>'  # big-endian

        data = np.fromfile(f, endian + 'f')
        shape = (height, width, 3) if color else (height, width)

        data = np.reshape(data, shape)
        data = np.flipud(data)

    # Mark invalid pixels as NaN
    mask = np.tile(data[:, :, 2:3], (1, 1, 2))
    flow = data[:, :, :2].astype(np.float32)
    flow[mask > 0.5] = float('nan')
    return flow



import torch
from torch import nn

BatchNorm = nn.BatchNorm2d


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        bias=False)


class BasicBlock(nn.Module):

    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            inplanes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation)
        self.bn1 = BatchNorm(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=1,
            padding=dilation,
            bias=False,
            dilation=dilation)
        self.bn2 = BatchNorm(planes)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(Bottleneck, self).__init__()
        expansion = Bottleneck.expansion
        bottle_planes = planes // expansion
        self.conv1 = nn.Conv2d(
            inplanes, bottle_planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm(bottle_planes)
        self.conv2 = nn.Conv2d(
            bottle_planes,
            bottle_planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation)
        self.bn2 = BatchNorm(bottle_planes)
        self.conv3 = nn.Conv2d(
            bottle_planes, planes, kernel_size=1, bias=False)
        self.bn3 = BatchNorm(planes)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += residual
        out = self.relu(out)

        return out


class BottleneckX(nn.Module):
    expansion = 2
    cardinality = 32

    def __init__(self, inplanes, planes, stride=1, dilation=1):
        super(BottleneckX, self).__init__()
        cardinality = BottleneckX.cardinality
        # dim = int(math.floor(planes * (BottleneckV5.expansion / 64.0)))
        # bottle_planes = dim * cardinality
        bottle_planes = planes * cardinality // 32
        self.conv1 = nn.Conv2d(
            inplanes, bottle_planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm(bottle_planes)
        self.conv2 = nn.Conv2d(
            bottle_planes,
            bottle_planes,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
            groups=cardinality)
        self.bn2 = BatchNorm(bottle_planes)
        self.conv3 = nn.Conv2d(
            bottle_planes, planes, kernel_size=1, bias=False)
        self.bn3 = BatchNorm(planes)
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride

    def forward(self, x, residual=None):
        if residual is None:
            residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        out += residual
        out = self.relu(out)

        return out


class Root(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size, residual):
        super(Root, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            1,
            stride=1,
            bias=False,
            padding=(kernel_size - 1) // 2)
        self.bn = BatchNorm(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.residual = residual

    def forward(self, *x):
        children = x
        x = self.conv(torch.cat(x, 1))
        x = self.bn(x)
        if self.residual:
            x += children[0]
        x = self.relu(x)

        return x


class Tree(nn.Module):

    def __init__(self,
                 levels,
                 block,
                 in_channels,
                 out_channels,
                 stride=1,
                 level_root=False,
                 root_dim=0,
                 root_kernel_size=1,
                 dilation=1,
                 root_residual=False):
        super(Tree, self).__init__()
        if root_dim == 0:
            root_dim = 2 * out_channels
        if level_root:
            root_dim += in_channels
        if levels == 1:
            self.tree1 = block(
                in_channels, out_channels, stride, dilation=dilation)
            self.tree2 = block(
                out_channels, out_channels, 1, dilation=dilation)
        else:
            self.tree1 = Tree(
                levels - 1,
                block,
                in_channels,
                out_channels,
                stride,
                root_dim=0,
                root_kernel_size=root_kernel_size,
                dilation=dilation,
                root_residual=root_residual)
            self.tree2 = Tree(
                levels - 1,
                block,
                out_channels,
                out_channels,
                root_dim=root_dim + out_channels,
                root_kernel_size=root_kernel_size,
                dilation=dilation,
                root_residual=root_residual)
        if levels == 1:
            self.root = Root(root_dim, out_channels, root_kernel_size,
                             root_residual)
        self.level_root = level_root
        self.root_dim = root_dim
        self.downsample = None
        self.project = None
        self.levels = levels
        if stride > 1:
            self.downsample = nn.MaxPool2d(stride, stride=stride)
        if in_channels != out_channels:
            self.project = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    stride=1,
                    bias=False), BatchNorm(out_channels))

    def forward(self, x, residual=None, children=None):
        children = [] if children is None else children
        bottom = self.downsample(x) if self.downsample else x
        residual = self.project(bottom) if self.project else bottom
        if self.level_root:
            children.append(bottom)
        x1 = self.tree1(x, residual)
        if self.levels == 1:
            x2 = self.tree2(x1)
            x = self.root(x2, x1, *children)
        else:
            children.append(x1)
            x = self.tree2(x1, children=children)
        return x


class DLA(nn.Module):

    def __init__(self,
                 levels,
                 channels,
                 num_classes=1000,
                 block=BasicBlock,
                 residual_root=False,
                 return_levels=False,
                 pool_size=7,
                 linear_root=False):
        super(DLA, self).__init__()
        self.channels = channels
        self.return_levels = return_levels
        self.num_classes = num_classes
        self.base_layer = nn.Sequential(
            nn.Conv2d(
                3, channels[0], kernel_size=7, stride=1, padding=3,
                bias=False), BatchNorm(channels[0]), nn.ReLU(inplace=True))
        self.level0 = self._make_conv_level(channels[0], channels[0],
                                            levels[0])
        self.level1 = self._make_conv_level(
            channels[0], channels[1], levels[1], stride=2)
        self.level2 = Tree(
            levels[2],
            block,
            channels[1],
            channels[2],
            2,
            level_root=False,
            root_residual=residual_root)
        self.level3 = Tree(
            levels[3],
            block,
            channels[2],
            channels[3],
            2,
            level_root=True,
            root_residual=residual_root)
        self.level4 = Tree(
            levels[4],
            block,
            channels[3],
            channels[4],
            2,
            level_root=True,
            root_residual=residual_root)
        self.level5 = Tree(
            levels[5],
            block,
            channels[4],
            channels[5],
            2,
            level_root=True,
            root_residual=residual_root)
        self.level6 = Tree(
            levels[6],
            block,
            channels[5],
            channels[6],
            2,
            level_root=True,
            root_residual=residual_root)

        for m in self.modules():
            classname = m.__class__.__name__
            if isinstance(m, nn.Conv2d):
                m.weight.data.normal_(0, 0.02)
            elif classname.find('BatchNorm') != -1:
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_conv_level(self, inplanes, planes, convs, stride=1, dilation=1):
        modules = []
        for i in range(convs):
            modules.extend([
                nn.Conv2d(
                    inplanes,
                    planes,
                    kernel_size=3,
                    stride=stride if i == 0 else 1,
                    padding=dilation,
                    bias=False,
                    dilation=dilation),
                BatchNorm(planes),
                nn.ReLU(inplace=True)
            ])
            inplanes = planes
        return nn.Sequential(*modules)

    def forward(self, x):
        y = []
        x = self.base_layer(x)
        for i in range(7):
            x = getattr(self, 'level{}'.format(i))(x)
            y.append(x)
        return y


def dla34(planes, **kwargs):
    assert len(planes) == 7
    model = DLA([1, 1, 1, 2, 2, 1, 1],
                planes,
                block=BasicBlock,
                return_levels=True,
                **kwargs)
    return model



from __future__ import absolute_import, division, print_function

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as tf


def occ_to_mask(occ, return_np=False):
    tens_out = nn.Sigmoid()(occ)
    if return_np:
        return np.round(tens_out.expand(-1, 3, -1, -1).data.cpu().numpy().transpose(
            [0, 2, 3, 1])) * 255
    return tens_out.round()


def _elementwise_epe(input_flow, target_flow):
    residual = target_flow - input_flow
    return torch.norm(residual, p=2, dim=1, keepdim=True)


def _elementwise_robust_epe_char(input_flow, target_flow):
    residual = target_flow - input_flow
    return torch.pow(torch.norm(residual, p=2, dim=1, keepdim=True) + 0.01, 0.4)


def _downsample2d_as(inputs, target_as):
    _, _, h, w = target_as.size()
    return tf.adaptive_avg_pool2d(inputs, [h, w])


def _upsample2d_as(inputs, target_as, mode="bilinear"):
    _, _, h, w = target_as.size()
    return tf.interpolate(inputs, [h, w], mode=mode, align_corners=True)


def f1_score(y_true, y_pred):
    return fbeta_score(y_true, y_pred, 1)


def fbeta_score(y_true, y_pred, beta, eps=1e-8):
    beta2 = beta ** 2

    y_pred = y_pred.float()
    y_true = y_true.float()

    true_positive = (y_pred * y_true).sum(dim=2).sum(dim=2)
    precision = true_positive / (y_pred.sum(dim=2).sum(dim=2) + eps)
    recall = true_positive / (y_true.sum(dim=2).sum(dim=2) + eps)

    return torch.mean(precision * recall / (precision * beta2 + recall + eps) * (1 + beta2))


def f1_score_bal_loss(y_pred, y_true):
    eps = 1e-8

    tp = -(y_true * torch.log(y_pred + eps)).sum(dim=2).sum(dim=2).sum(dim=1)
    fn = -((1 - y_true) * torch.log((1 - y_pred) + eps)).sum(dim=2).sum(dim=2).sum(dim=1)

    denom_tp = y_true.sum(dim=2).sum(dim=2).sum(dim=1) + y_pred.sum(dim=2).sum(dim=2).sum(dim=1) + eps
    denom_fn = (1 - y_true).sum(dim=2).sum(dim=2).sum(dim=1) + (1 - y_pred).sum(dim=2).sum(dim=2).sum(dim=1) + eps

    return ((tp / denom_tp).sum() + (fn / denom_fn).sum()) * y_pred.size(2) * y_pred.size(3) * 0.5


class MultiScaleEPE_PWC_Bi_Occ_upsample(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005, 0.00125, 0.0003125]

        self.occ_activ = nn.Sigmoid()
        self.f1_score_bal_loss = f1_score_bal_loss

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_flo_b = self._args.div_flow * target_dict["flows_b"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]
            target_occ_b = target_dict["occs_b"][:, 0]

            # bchw
            flow_loss = 0
            occ_loss = 0

            for ii, output_ii in enumerate(output_flo):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    loss_ii = loss_ii + _elementwise_epe(output_ii[2 * jj],
                                                         _downsample2d_as(target_flo_f, output_ii[2 * jj])).sum()
                    loss_ii = loss_ii + _elementwise_epe(output_ii[2 * jj + 1],
                                                         _downsample2d_as(target_flo_b, output_ii[2 * jj + 1])).sum()
                flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii)

            for ii, output_ii in enumerate(output_occ):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    output_occ_f = self.occ_activ(output_ii[2 * jj])
                    output_occ_b = self.occ_activ(output_ii[2 * jj + 1])
                    loss_ii = loss_ii + self.f1_score_bal_loss(output_occ_f,
                                                               _downsample2d_as(target_occ_f, output_occ_f))
                    loss_ii = loss_ii + self.f1_score_bal_loss(output_occ_b,
                                                               _downsample2d_as(target_occ_b, output_occ_b))
                occ_loss = occ_loss + self._weights[ii] * loss_ii / len(output_ii)

            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if (f_loss.data > o_loss.data).numpy:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["occ_loss"] = occ_loss / self._batch_size
            loss_dict["loss"] = (flow_loss * f_l_w + occ_loss * o_l_w) / self._batch_size

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flow"], target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0], torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict


class MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel(nn.Module):
    def __init__(self, args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.32, 0.08, 0.02, 0.01, 0.005, 0.00125, 0.0003125]

        self.occ_activ = nn.Sigmoid()
        self.occ_loss_bce = nn.BCELoss(reduction='sum')
        self.seploss = hasattr(args, 'seploss') and self._args.seploss or False
        if self.seploss:
            print("Starting MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel loss with seploss!")
        self.loss_perc = hasattr(args, 'loss_perc') and args.loss_perc
        if self.loss_perc:
            from torchpercentile import Percentile
            # from tensorboard import summary
            # self.writer = summary(args.save)
            from matplotlib.pyplot import hist
            self.perc = Percentile()
            print("Starting MultiScaleEPE_PWC_Bi_Occ_upsample_Sintel loss percentile!")
            self.min_p = nn.Parameter(torch.Tensor([30]), requires_grad=False)
            self.max_p = nn.Parameter(torch.Tensor([97]), requires_grad=False)

    def _get_flow_loss(self, output_flo, target_flo_f):
        flow_loss = 0
        for ii, output_ii in enumerate(output_flo):
            loss_ii = 0
            for jj in range(0, len(output_ii) // 2):
                cur_epe = _elementwise_robust_epe_char(output_ii[2 * jj],
                                                       _downsample2d_as(target_flo_f,
                                                                        output_ii[2 * jj]))
                if self.loss_perc:
                    tmin = torch.Tensor([self.perc(ten.flatten(), [self.min_p])
                                         for ten in cur_epe])
                    # tmin.requires_grad = False
                    tmax = torch.Tensor([self.perc(ten.flatten(), [self.max_p])
                                         for ten in cur_epe])
                    # tmax.requires_grad = False
                    cur_epe[cur_epe > tmax.view(-1, 1, 1, 1).cuda()] = 0.
                    cur_epe[cur_epe < tmin.view(-1, 1, 1, 1).cuda()] = 0.
                    tmin.detach()
                    tmax.detach()
                # self.writer.add_histogram('flow_epe', cur_epe.grad)
                # self.writer.add_scalar('test', 1)
                loss_ii = loss_ii + cur_epe.sum()
                output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
            flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii) * 2
        return flow_loss

    def _get_occ_loss(self, output_occ, target_occ_f):
        occ_loss = 0
        for ii, output_ii in enumerate(output_occ):
            loss_ii = 0
            for jj in range(0, len(output_ii) // 2):
                output_occ_f = self.occ_activ(output_ii[2 * jj])
                output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
                loss_ii = loss_ii + self.occ_loss_bce(output_occ_f,
                                                      _downsample2d_as(target_occ_f,
                                                                       output_occ_f))
            occ_loss = occ_loss + self._weights[ii] * loss_ii / len(output_ii) * 2
        return occ_loss

    @staticmethod
    def mask_flow(output_flo, target_flo_f, output_occ, use_target=False, min_layer=0):
        for lix, layer_flow in enumerate(output_flo):
            if lix < min_layer:
                continue

            for fcix, flow_category in enumerate(layer_flow):
                if use_target:
                    occ_mask = _downsample2d_as(target_flo_f, output_flo[lix][fcix])
                else:
                    occ_mask = occ_to_mask(output_occ[lix][fcix])
                output_flo[lix][fcix] = (1 - occ_mask) * output_flo[lix][
                    fcix]

        if use_target:
            target_flo_f = (1 - target_flo_f) * target_flo_f
        else:
            target_flo_f = (1 - occ_to_mask(output_occ[6][0])) * target_flo_f

        return output_flo, target_flo_f

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        if self.training:
            # Extract output
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # Extract and prepare target
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]
            target_occ_f = target_dict["occs"][:, 0]

            # Mask with occ
            if self.seploss:
                output_flo, target_flo_f = self.mask_flow(output_flo,
                                                          target_flo_f,
                                                          output_occ)

            # Calc multi level losses
            flow_loss = self._get_flow_loss(output_flo, target_flo_f)
            occ_loss = self._get_occ_loss(output_occ, target_occ_f)
            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["occ_loss"] = occ_loss / self._batch_size

            # Calc total loss
            f_loss = flow_loss.detach()
            o_loss = occ_loss.detach()
            if (f_loss.data > o_loss.data).numpy:
                f_l_w = 1
                o_l_w = f_loss / o_loss
            else:
                f_l_w = o_loss / f_loss
                o_l_w = 1
            losses_mix = (flow_loss * f_l_w + occ_loss * o_l_w)
            loss_dict["loss"] = losses_mix / self._batch_size

        else:
            loss_dict["epe"] = _elementwise_epe(output_dict["flow"],
                                                target_dict["flows"][:, 0]).mean()
            loss_dict["F1"] = f1_score(target_dict["occs"][:, 0],
                                       torch.round(self.occ_activ(output_dict["occ"])))

        return loss_dict


class MultiScaleEPE_PWC_Bi_Occ_upsample_KITTI(nn.Module):
    def __init__(self,
                 args):

        super(MultiScaleEPE_PWC_Bi_Occ_upsample_KITTI, self).__init__()
        self._args = args
        self._batch_size = args.train_batch_size
        self._weights = [0.001, 0.001, 0.001, 0.002, 0.004, 0.004, 0.004]

        self.occ_activ = nn.Sigmoid()

    def forward(self, output_dict, target_dict):
        loss_dict = {}

        valid_mask = target_dict["valids"][:, 0]
        b, _, h, w = target_dict["flows"][:, 0].size()

        if self.training:
            output_flo = output_dict['flow_preds']
            output_occ = output_dict['occ_preds']

            # div_flow trick
            target_flo_f = self._args.div_flow * target_dict["flows"][:, 0]

            # bchw
            flow_loss = 0

            for ii, output_ii in enumerate(output_flo):
                loss_ii = 0
                for jj in range(0, len(output_ii) // 2):
                    valid_epe = _elementwise_robust_epe_char(_upsample2d_as(output_ii[2 * jj], target_flo_f),
                                                             target_flo_f) * valid_mask

                    for bb in range(0, b):
                        valid_epe[bb, ...][valid_mask[bb, ...] == 0] = valid_epe[bb, ...][
                            valid_mask[bb, ...] == 0].detach()
                        norm_const = h * w / (valid_mask[bb, ...].sum())
                        loss_ii = loss_ii + valid_epe[bb, ...][valid_mask[bb, ...] != 0].sum() * norm_const

                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()
                flow_loss = flow_loss + self._weights[ii] * loss_ii / len(output_ii) * 2

            for ii, output_ii in enumerate(output_occ):
                for jj in range(0, len(output_ii) // 2):
                    output_ii[2 * jj] = output_ii[2 * jj].detach()
                    output_ii[2 * jj + 1] = output_ii[2 * jj + 1].detach()

            loss_dict["flow_loss"] = flow_loss / self._batch_size
            loss_dict["loss"] = flow_loss / self._batch_size

        else:
            flow_gt_mag = torch.norm(target_dict["flows"][:, 0], p=2, dim=1, keepdim=True) + 1e-8
            flow_epe = _elementwise_epe(output_dict["flow"], target_dict["flows"][:, 0]) * valid_mask

            epe_per_image = (flow_epe.view(b, -1).sum(1)) / (valid_mask.view(b, -1).sum(1))
            loss_dict["epe"] = epe_per_image.mean()

            outlier_epe = (flow_epe > 3).float() * ((flow_epe / flow_gt_mag) > 0.05).float() * valid_mask
            outlier_per_image = (outlier_epe.view(b, -1).sum(1)) / (valid_mask.view(b, -1).sum(1))
            loss_dict["outlier"] = outlier_per_image.mean()

        return loss_dict


class NO_OP(nn.Module):
    def __init__(self, args=None):
        super(NO_OP, self).__init__()

    def forward(self, output_dict, target_dict):
        return {'flow_loss': -1, 'epe': -1,
                'total_loss': torch.Tensor([0])}


from argparse import Namespace
from pathlib import Path

import torch
import torch.nn as nn
from torch.nn import init

from .flownetc import FlowNetC
from .flownets import FlowNetS
from .flownetsd import FlowNetSD
from .flownet_fusion import FlowNetFusion
from .submodules import *
from .flownet_base import FlowNetBase

class FlowNet2(FlowNetBase):
    pretrained_checkpoints = {
        'things': 'https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flownet2-things-d63b53a7.ckpt'
    }
    
    def __init__(self, args: Namespace):
        args.input_channels = 12
        args.loss_start_scale = 1
        args.loss_num_scales = 3
        super(FlowNet2, self).__init__(args)

        self.args = args
        self.rgb_max = 1

        # First Block (FlowNetC)
        self.flownetc = FlowNetC(args)

        # Block (FlowNetS)
        self.flownets_1 = FlowNetS(args)
        self.flownets_2 = FlowNetS(args)

        # Block (FlowNetSD)
        self.flownets_d = FlowNetSD(args)

        # Block (FLowNetFusion)
        self.flownetfusion = FlowNetFusion(args)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)

    def warp(self, x, flo):
        """
        warp an image/tensor (im2) back to im1, according to the optical flow
        x: [B, C, H, W] (im2)
        flo: [B, 2, H, W] flow
        """
        B, C, H, W = x.size()
        # mesh grid 
        xx = torch.arange(0, W).view(1,-1).repeat(H,1)
        yy = torch.arange(0, H).view(-1,1).repeat(1,W)
        xx = xx.view(1,1,H,W).repeat(B,1,1,1)
        yy = yy.view(1,1,H,W).repeat(B,1,1,1)
        grid = torch.cat((xx,yy),1).float()

        if x.is_cuda:
            grid = grid.to(dtype=x.dtype, device=x.device)
        vgrid = grid + flo

        # scale grid to [-1,1] 
        vgrid[:,0,:,:] = 2.0*vgrid[:,0,:,:].clone() / max(W-1,1)-1.0
        vgrid[:,1,:,:] = 2.0*vgrid[:,1,:,:].clone() / max(H-1,1)-1.0

        vgrid = vgrid.permute(0,2,3,1)        
        output = nn.functional.grid_sample(x, vgrid, align_corners=True)
        mask = torch.ones(x.size()).to(dtype=x.dtype, device=x.device)
        mask = nn.functional.grid_sample(mask, vgrid, align_corners=True)
        
        mask[mask<0.9999] = 0
        mask[mask>0] = 1
        
        return output*mask

    def normalize(self, im):
        im = im - 0.5
        im = im / 0.5
        return im

    def forward(self, inputs):
        # flownetc
        flownetc_flow = self.flownetc(inputs)['flows'][:, 0]
        
        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.warp(inputs['images'][:, 1], flownetc_flow)
        diff_img0 = inputs['images'][:, 0] - resampled_img1 
        norm_diff_img0 = torch.norm(diff_img0, p=2, dim=1, keepdim=True)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((inputs['images'][:, 0], inputs['images'][:, 1], resampled_img1, flownetc_flow/self.args.div_flow, norm_diff_img0), dim=1)
        
        # flownets1
        flownets1_flow = self.flownets_1({'images': concat1[:, None]})['flows'][:, 0]

        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.warp(inputs['images'][:, 1], flownets1_flow)
        diff_img0 = inputs['images'][:, 0] - resampled_img1 
        norm_diff_img0 = torch.norm(diff_img0, p=2, dim=1, keepdim=True)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat2 = torch.cat((inputs['images'][:, 0], inputs['images'][:, 1], resampled_img1, flownets1_flow/self.args.div_flow, norm_diff_img0), dim=1)
        
        # flownets2
        flownets2_flow = self.flownets_2({'images': concat2[:, None]})['flows'][:, 0]
        norm_flownets2_flow = torch.norm(flownets2_flow, p=2, dim=1, keepdim=True)

        diff_flownets2_flow = self.warp(inputs['images'][:, 1], flownets2_flow)
        diff_flownets2_img1 = torch.norm(inputs['images'][:, 0]-diff_flownets2_flow, p=2, dim=1, keepdim=True)

        # flownetsd
        flownetsd_flow = self.flownets_d(inputs)['flows'][:, 0] / self.args.div_flow**2
        norm_flownetsd_flow = torch.norm(flownetsd_flow, p=2, dim=1, keepdim=True)
        
        diff_flownetsd_flow = self.warp(inputs['images'][:, 1], flownetsd_flow)
        diff_flownetsd_img1 = torch.norm(inputs['images'][:, 0]-diff_flownetsd_flow, p=2, dim=1, keepdim=True)

        # concat img1 flownetsd, flownets2, norm_flownetsd, norm_flownets2, diff_flownetsd_img1, diff_flownets2_img1
        concat3 = torch.cat((inputs['images'][:, 0], flownetsd_flow, flownets2_flow, norm_flownetsd_flow, norm_flownets2_flow, diff_flownetsd_img1, diff_flownets2_img1), dim=1)
        flownetfusion_preds = self.flownetfusion({'images': concat3[:, None]})

        return flownetfusion_preds
